@@ -1,18 +1,35 @@
+from __future__ import annotations
+
 import functools
 import hashlib
-from dataclasses import dataclass
+import time
 from itertools import chain
 
 import gitlab.const
+from attrs import define
 from github import Auth, Consts, Github, GithubException, PullRequest
 from gitlab import Gitlab, GitlabAuthenticationError, GitlabError
 from gitlab.v4.objects import ProjectMergeRequest
-from typing_extensions import Protocol
+from typing_extensions import Protocol, TypedDict
 
 from patchwork.logger import logger
 
 
-@dataclass(slots=True)
+def get_slug_from_remote_url(remote_url: str) -> str:
+    # TODO: consider using https://github.com/nephila/giturlparse instead
+    if remote_url.startswith("git@"):
+        # ssh
+        _, _, potential_slug = remote_url.partition(":")
+    else:
+        potential_slug = "/".join(remote_url.split("/")[-2:])
+
+    if potential_slug.endswith(".git"):
+        potential_slug = potential_slug[:-4]
+
+    return potential_slug
+
+
+@define
 class Comment:
     path: str
     body: str
@@ -20,6 +37,16 @@ class Comment:
     end_line: int
 
 
+class IssueText(TypedDict):
+    title: str
+    body: str
+    comments: list[str]
+
+
+class PullRequestTexts(IssueText):
+    diffs: dict[str, str]
+
+
 _COMMENT_MARKER = "<!-- PatchWork comment marker -->"
 
 
@@ -38,7 +65,7 @@ class PullRequestProtocol(Protocol):
     def reset_comments(self) -> None:
         ...
 
-    def file_diffs(self) -> dict[str, str]:
+    def texts(self) -> PullRequestTexts:
         ...
 
     @staticmethod
@@ -108,10 +135,10 @@ class ScmPlatformClientProtocol(Protocol):
     def get_slug_and_id_from_url(self, url: str) -> tuple[str, int] | None:
         ...
 
-    def find_issue_by_url(self, url: str) -> list[str] | None:
+    def find_issue_by_url(self, url: str) -> IssueText | None:
         ...
 
-    def find_issue_by_id(self, slug: str, issue_id: int) -> list[str] | None:
+    def find_issue_by_id(self, slug: str, issue_id: int) -> IssueText | None:
         ...
 
     def get_pr_by_url(self, url: str) -> PullRequestProtocol | None:
@@ -158,23 +185,29 @@ class GitlabMergeRequest(PullRequestProtocol):
             note = self._mr.notes.create({"body": final_body})
             return f"#note_{note.get_id()}"
 
-        while True:
+        commit = None
+        for i in range(3):
             try:
                 commit = self._mr.commits().next()
             except StopIteration:
-                continue
+                time.sleep(2**i)
 
-            break
+        if commit is None:
+            return None
 
-        while True:
+        diff = None
+        for i in range(3):
             try:
                 iterator = self._mr.diffs.list(iterator=True)
                 diff = iterator.next()  # type: ignore
+                if diff.head_commit_sha == commit.get_id():
+                    break
             except StopIteration:
+                time.sleep(2**i)
                 continue
 
-            if diff.head_commit_sha == commit.get_id():
-                break
+        if diff is None:
+            return None
 
         base_commit = diff.base_commit_sha
         head_commit = diff.head_commit_sha
@@ -210,14 +243,18 @@ class GitlabMergeRequest(PullRequestProtocol):
                 if note["body"].startswith(_COMMENT_MARKER):
                     discussion.notes.delete(note["id"])
 
-    def file_diffs(self) -> dict[str, str]:
+    def texts(self) -> PullRequestTexts:
+        title = self._mr.title
+        body = self._mr.description
+        notes = [note.body for note in self._mr.notes.list() if note.system is False]
+
         diffs = self._mr.diffs.list()
         latest_diff = max(diffs, key=lambda diff: diff.created_at, default=None)
         if latest_diff is None:
-            return {}
+            return dict(title=title, body=body, comments=notes, diffs={})
 
         files = self._mr.diffs.get(latest_diff.id).diffs
-        return {file["new_path"]: file["diff"] for file in files}
+        return dict(title=title, body=body, comments=notes, diffs={file["new_path"]: file["diff"] for file in files})
 
 
 class GithubPullRequest(PullRequestProtocol):
@@ -254,9 +291,13 @@ class GithubPullRequest(PullRequestProtocol):
             if comment.body.startswith(_COMMENT_MARKER):
                 comment.delete()
 
-    def file_diffs(self) -> dict[str, str]:
-        files = self._pr.get_files()
-        return {file.filename: file.patch for file in files}
+    def texts(self) -> PullRequestTexts:
+        return dict(
+            title=self._pr.title or "",
+            body=self._pr.body or "",
+            comments=[comment.body for comment in self._pr.get_comments()],
+            diffs={file.filename: file.patch for file in self._pr.get_files()},
+        )
 
 
 class GithubClient(ScmPlatformClientProtocol):
@@ -293,17 +334,19 @@ class GithubClient(ScmPlatformClientProtocol):
 
         return slug, resource_id
 
-    def find_issue_by_url(self, url: str) -> list[str] | None:
+    def find_issue_by_url(self, url: str) -> IssueText | None:
         slug, issue_id = self.get_slug_and_id_from_url(url)
         return self.find_issue_by_id(slug, issue_id)
 
-    def find_issue_by_id(self, slug: str, issue_id: int) -> list[str] | None:
+    def find_issue_by_id(self, slug: str, issue_id: int) -> IssueText | None:
         repo = self.github.get_repo(slug)
         try:
             issue = repo.get_issue(issue_id)
-            body = issue.body
-            comments = [issue_comment.body for issue_comment in issue.get_comments()]
-            return [body] + comments
+            return dict(
+                title=issue.title,
+                body=issue.body,
+                comments=[issue_comment.body for issue_comment in issue.get_comments()],
+            )
         except GithubException as e:
             logger.warn(f"Failed to get issue: {e}")
             return None
@@ -391,17 +434,19 @@ class GitlabClient(ScmPlatformClientProtocol):
 
         return slug, resource_id
 
-    def find_issue_by_url(self, url: str) -> list[str] | None:
+    def find_issue_by_url(self, url: str) -> IssueText | None:
         slug, issue_id = self.get_slug_and_id_from_url(url)
         return self.find_issue_by_id(slug, issue_id)
 
-    def find_issue_by_id(self, slug: str, issue_id: int) -> list[str] | None:
+    def find_issue_by_id(self, slug: str, issue_id: int) -> IssueText | None:
         project = self.gitlab.projects.get(slug)
         try:
             issue = project.issues.get(issue_id)
-            body = issue["description"]
-            comments = [note["body"] for note in issue.notes.list()]
-            return [body] + comments
+            return dict(
+                title=issue.get("title", ""),
+                body=issue.get("description", ""),
+                comments=[note["body"] for note in issue.notes.list()],
+            )
         except GitlabError as e:
             logger.warn(f"Failed to get issue: {e}")
             return None
@@ -452,6 +497,7 @@ class GitlabClient(ScmPlatformClientProtocol):
                 "target_branch": original_branch,
                 "title": title,
                 "description": body,
+                "labels": "patchwork",
             }
         )
         mr = GitlabMergeRequest(gl_mr)  # type: ignore
